import logging
import decimal
import inspect
import sqlite3
import pmg
from . import factories

DEFAULT_PRAGMAS = {
    'temp_store': 2,
    'foreign_keys': 1
}

namedtuple_factory = factories.namedtuple_factory
log = logging.getLogger(__name__)

def open_db(path=None, pragmas=None, read_only=False, timeout=10, **kwargs):
    cn = None
    path = pmg.coalesce(path, ':memory:')
    if read_only:
        cn = sqlite3.connect(f"file:{path}?mode=ro", timeout=timeout, **kwargs)
    else:
        cn = sqlite3.connect(path, timeout=timeout, **kwargs)
        if pragmas and 'mode=ro' not in path:
            cur = cn.cursor()
            for k, v in pragmas.items():
                cur.execute(f'pragma {k}="{v}";')
            cn.commit()
    return cn

def open_db_ro(path, timeout=10, **kwargs):
    return open_db(path, read_only=True, timeout=timeout, **kwargs)

def adapt_decimal(d):
    return str(d)

def backup_progress(status, remaining, total):
    log.debug('SQLite backup copied %d of %d pages [status: %s].', total - remaining, total, status)

def backup(target_path: str, source_db, pages=-1):
    with open_db(target_path) as dbTarget:
        source_db.backup(dbTarget, pages=pages, progress=backup_progress)
        dbTarget.commit()

def add_functions(cn, *functions):
    for func in functions:
        cn.create_function(func.__name__, len(inspect.getfullargspec(func)[0]), func)


sqlite3.register_adapter(decimal.Decimal, adapt_decimal)



import functools
import logging
import platform
import os
import sched
import time
import pmg


log = logging.getLogger(__name__)

def get_os_name():
    os_name = f"{platform.system()} {platform.version()}"
    try:
        if os.path.isfile('/etc/os-release'):
            with open('/etc/os-release', 'rt', encoding='utf-8') as f:
                for line in f:
                    k, v = line.split('=')
                    if k == "PRETTY_NAME":
                        os_name = pmg.unquote(v.strip(), '"')
                        break
    except Exception:
        pass
    return f'{os_name} ({platform.platform()})'

def get_package_versions(*packages_start_with):
    versions = []
    import pkg_resources
    for pkg in iter(pkg_resources.working_set):
        dist = pkg_resources.get_distribution(pkg)
        if any([dist.project_name.startswith(package_name) for package_name in packages_start_with]):
            versions.append(f'{dist.project_name} {dist.version}')
    return sorted(versions)

def run_every(sleep_seconds, run_at_start=False, wait_after_fail=15):
    def wrapper(func):
        @functools.wraps(func)
        def looping_function(*args, **kwargs):
            @pmg.logwrap(log)
            def run_it():
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    log.exception('Exception during service function, pausing %d seconds...', wait_after_fail)
                    time.sleep(wait_after_fail)
            if run_at_start:
                run_it()
            sch = sched.scheduler()
            while True:
                sch.enter(sleep_seconds, 1, run_it)
                sch.run()
        return looping_function
    return wrapper


import builtins
import datetime
import decimal
import numbers
import pmg


class UnboxException(Exception):
    pass

def unbox_bool(value):
    if value is None:
        return value
    if isinstance(value, bool):
        return value
    if isinstance(value, (numbers.Number, decimal.Decimal)):
        return False if value == 0 else True
    if isinstance(value, str):
        return False if value.upper() in ['0', 'FALSE', 'N', ''] else True
    raise UnboxException(f"Cannot unbox value '{value}' from type {type(value)} to bool.")

def unbox_date(value):
    if value is None:
        return value
    if type(value) == datetime.datetime:
        return value.date()
    if type(value) == datetime.date:
        return value
    if type(value) == str:
        # Expect ISO-formatted date string
        return datetime.date(int(value[:4]), int(value[5:7]), int(value[8:10]))
    raise UnboxException(f"Cannot unbox value '{value}' from type {type(value)} to date.")

def unbox_datetime(value):
    if value is None:
        return value
    if type(value) == datetime.datetime:
        return value.date()
    if type(value) == str:
        # Expect ISO-formatted date string
        return pmg.parse_iso_datetime_utc(value)
    raise UnboxException(f"Cannot unbox value '{value}' from type {type(value)} to datetime.")

def unbox_int(value):
    if value is None:
        return value
    if isinstance(value, numbers.Integral):
        return value
    if isinstance(value, (numbers.Real, decimal.Decimal)):
        return int(value)
    if isinstance(value, str):
        return int(float(value)) if '.' in value else int(value)
    raise UnboxException(f"Cannot unbox value '{value}' from type {type(value)} to int.")

def unbox_float(value):
    if value is None:
        return value
    if isinstance(value, numbers.Integral):
        return float(value)
    if isinstance(value, (numbers.Real, decimal.Decimal)):
        return float(value)
    if isinstance(value, str):
        return float(value)
    raise UnboxException(f"Cannot unbox value '{value}' from type {type(value)} to float.")

def unbox_decimal(value):
    if value is None:
        return value
    if isinstance(value, decimal.Decimal):
        return value
    if isinstance(value, (numbers.Real, numbers.Integral, str)):
        return decimal.Decimal(value)
    raise UnboxException(f"Cannot unbox value '{value}' from type {type(value)} to decimal.")

def unbox_str(value):
    if value is None:
        return value
    if isinstance(value, str):
        return value
    raise UnboxException(f"Cannot unbox value '{value}' from type {type(value)} to str.")

FUNCTIONS = {
    'bool': unbox_bool,
    'int': unbox_int,
    'Decimal': unbox_decimal,
    'date': unbox_date,
    'datetime': unbox_datetime,
    'str': unbox_str
}

def box(o):
    try:
        if isinstance(o, decimal.Decimal):
            return format(o, 'f')
        if isinstance(o, (datetime.date, datetime.time)):
            return o.isoformat()
        return o
    except Exception as ex:
        return f'<Unencodable type {builtins.type(o)} ({type(ex)})>'

def unbox(value, ty):
    return FUNCTIONS[ty](value)



"""Tests for the pmg.base36 module"""

from hypothesis import given, example
import hypothesis.strategies as st
from pmg.base36 import b36encode, b36decode

@given(st.integers())
@example(0)
def test_encode_decode_match(x):
    assert x == b36decode(b36encode(x))

@given(st.integers(min_value=1))
def test_pos_neg_match(x):
    assert b36encode(-x) == f'-{b36encode(x)}'



"""Base36 conversion functions"""

import numbers

ALPHABET = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'

def b36encode(number):
    """Converts an integer to a base36 string."""
    if not isinstance(number, numbers.Integral):
        raise TypeError('number must be an integer')

    base36 = ''
    sign = ''

    if number < 0:
        sign = '-'
        number = -number

    if 0 <= number < len(ALPHABET):
        return sign + ALPHABET[number]

    while number != 0:
        number, i = divmod(number, len(ALPHABET))
        base36 = ALPHABET[i] + base36

    return sign + base36

def b36decode(number):
    """Converts a base36 integer to a base36 string."""
    return int(number, 36)



import collections
import functools

@functools.lru_cache(maxsize=64)
def make_namedtuple(cur_desc):
    return collections.namedtuple("Row", [col[0] for col in cur_desc])

def namedtuple_factory(cursor, row):
    """Returns sqlite rows as named tuples."""
    Row = make_namedtuple(cursor.description)
    return Row(*row)

def dict_factory(cursor, row):
    return dict(zip([col[0] for col in cursor.description], row))



"""Tests for the pmg module"""

import datetime
from hypothesis import given, example
import hypothesis.strategies as st
import pytest
import pmg

def test_get_next_weekday():
    today = datetime.datetime.today().date()
    tomorrow = today + datetime.timedelta(days=1)
    max_result = today + datetime.timedelta(days=7)
    for i in range(7):
        dt = pmg.get_next_weekday(i)
        assert dt.weekday() == i
        assert dt >= today and dt < max_result
        dt = pmg.get_next_weekday(i, from_date=tomorrow)
        assert dt.weekday() == i
        assert dt > today and dt <= max_result

@given(st.integers(min_value=0, max_value=6), st.dates())
def test_get_next_weekday_2(i, dt):
    res = pmg.get_next_weekday(i, dt)
    assert res >= dt and res < dt + datetime.timedelta(days=7)
    assert res.weekday() == i

def test_parse_iso_datetime():
    assert pmg.parse_iso_datetime_utc("1999-12-31 23:59:59.499999") == datetime.datetime(1999, 12, 31, 23, 59, 59, 499999, datetime.timezone.utc)
    assert pmg.parse_iso_datetime_utc("1985-10-26 01:21:42") == datetime.datetime(1985, 10, 26, 1, 21, 42, tzinfo=datetime.timezone.utc)

@given(st.datetimes(timezones=st.timezones()))
@example(pmg.utcnow())
def test_parse_iso_datetimes(dt):
    dt = dt.astimezone(datetime.timezone.utc)
    assert pmg.parse_iso_datetime_utc(dt.isoformat()) == dt

@given(st.dates())
def test_parse_iso_date(dt):
    assert pmg.parse_iso_date(dt.isoformat()) == dt

@given(st.datetimes(timezones=st.none()))
def test_is_datetime_naive(dt):
    assert pmg.is_datetime_naive(dt)

@given(st.datetimes(timezones=st.timezones()))
def test_is_datetime_not_naive(dt):
    assert not pmg.is_datetime_naive(dt)

@given(st.datetimes(timezones=st.timezones()))
def test_make_datetime_naive(dt):
    d = pmg.make_datetime_naive(dt)
    assert pmg.is_datetime_naive(d)
    assert d.year == dt.year and d.month == dt.month and d.day == dt.day and d.hour == dt.hour and d.minute == dt.minute and d.second == dt.second

@given(st.text(), st.integers(min_value=0))
def test_hide_str_excess(text, max_len):
    result = pmg.hide_str_excess(text, max_len)
    assert len(result) <= max_len + 3
    if len(result) <= max_len:
        assert result == text
    else:
        assert result[-3:] == '...'

@given(st.datetimes(timezones=st.timezones()), st.integers(min_value=-86400 * 365 * 20, max_value=86400 * 365 * 20))
def test_get_seconds_until(dt, i):
    assert pmg.get_seconds_until(dt, dt) == 0
    dt2 = dt + datetime.timedelta(seconds=i)
    assert pmg.get_seconds_until(dt2, dt) == i
    assert pmg.get_seconds_until(dt, dt2) == -i

def inner_quote_test(s, q):
    if q not in s:
        assert pmg.quote(s, q) == f'{q}{s}{q}'
        assert pmg.unquote(pmg.quote(s, q), q) == s
        assert pmg.unquote(s, q) == s
    else:
        with pytest.raises(AssertionError):
            assert pmg.quote(s, q) == f'{q}{s}{q}'

@given(st.text(), st.text())
@example("This won't fly", "'")
@example('"This will not fly, either."', '"')
def test_quote(s, q):
    inner_quote_test(s, q)

@given(st.text())
@example("This won't fly")
def test_quote_single(s):
    inner_quote_test(s, "'")

@given(st.text())
@example('"This will not fly, either."')
def test_quote_double(s):
    inner_quote_test(s, '"')

def test_localnow():
    dt = pmg.localnow('America/Chicago')
    utc = dt.astimezone(datetime.timezone.utc)
    assert not pmg.is_datetime_naive(dt)
    assert dt.tzinfo.utcoffset(dt).total_seconds() in [-5 * 3600, -6 * 3600]
    assert pmg.make_datetime_naive(dt) != pmg.make_datetime_naive(utc)

    dt = pmg.localnow('Europe/Zurich')
    utc = dt.astimezone(datetime.timezone.utc)
    assert not pmg.is_datetime_naive(dt)
    assert dt.tzinfo.utcoffset(dt).total_seconds() in [1 * 3600, 2 * 3600]
    assert pmg.make_datetime_naive(dt) != pmg.make_datetime_naive(utc)

    dt = pmg.localnow('Europe/London')
    utc = dt.astimezone(datetime.timezone.utc)
    assert not pmg.is_datetime_naive(dt)
    assert dt.tzinfo.utcoffset(dt).total_seconds() in [0, 3600]

@given(st.text(), st.text())
@example("ABCDEF", "AB")
@example("ABCDEF", "BA")
def test_remove_prefix(text, prefix):
    if text.startswith(prefix):
        assert pmg.remove_prefix(text, prefix) == text[len(prefix):]
    else:
        assert pmg.remove_prefix(text, prefix) == text

@given(st.text(), st.text())
@example("ABCDEF", "EF")
@example("ABCDEF", "BA")
def test_remove_suffix(text, suffix):
    if text.endswith(suffix):
        assert pmg.remove_suffix(text, suffix) == text[-len(suffix):]
    else:
        assert pmg.remove_suffix(text, suffix) == text



import logging
import time
import pmg
import pmg.sqlite

log = logging.getLogger(__name__)


class SqliteQueue:
    """
        The minimal table needed for a queue:
        CREATE TABLE "<queue_name>" (
        "<task_name>Id"	INTEGER NOT NULL,
        "StartDt"	TEXT,
        "EndDt"	TEXT,
        "Runtime"	INTEGER,
        "Status"	TEXT,
        "Result"	TEXT,
        "LastUpdateDt"	TEXT,
        PRIMARY KEY("<task_name>Id" AUTOINCREMENT)
    );
    """
    def __init__(self, db_path, queue_name, task_name=None, reset_running_tasks=False):
        self.queue_name = queue_name
        self.task_name = task_name or 'Job'
        self.db = pmg.sqlite.open_db(db_path)
        self.db.row_factory = pmg.sqlite.namedtuple_factory
        if reset_running_tasks:
            # Reset tests flagged as RUNNING
            self.db.execute(f"UPDATE {self.queue_name} SET Status = NULL, StartDt = NULL WHERE Status = 'RUNNING'")
            self.db.commit()

    def get_work(self, sleep_interval, limit_per_run):
        while True:
            work = self.db.execute(f"SELECT {self.task_name}Id AS QueuedTaskId, * FROM {self.queue_name} WHERE Status IS NULL ORDER BY {self.task_name}Id LIMIT ?", (limit_per_run, )).fetchall()
            for r in work:
                start_time = time.time()
                self.update_work_item(r.QueuedTaskId, 'RUNNING')
                yield r
                run_time = time.time() - start_time
                self.db.execute(f"UPDATE {self.queue_name} SET Runtime = ? WHERE {self.task_name}Id = ?", (run_time, r.QueuedTaskId))
                self.db.commit()
                log.debug("Task #%d in Queue '%s' completed after %f seconds.", r.QueuedTaskId, self.queue_name, run_time, extra={f"{self.task_name}Id": r.QueuedTaskId, "Queue": self.queue_name})
            time.sleep(sleep_interval)

    def run_queue(self, proc_func, *args, **kwargs):
        for r in self.get_work(*args, **kwargs):
            try:
                result = proc_func(r)
                self.update_work_item(r.QueuedTaskId, 'OK', str(result))
            except Exception as e:
                log.exception("Task #%d in Queue '%s' failed with exception:\n%s", r.QueuedTaskId, self.queue_name, str(e), extra={f"{self.task_name}Id": r.QueuedTaskId, "Queue": self.queue_name})
                self.update_work_item(r.QueuedTaskId, 'ERROR', str(e))

    def update_work_item(self, work_item_id, status, result=None):
        if status in ['OK', 'ERROR']:
            self.db.execute(f"UPDATE {self.queue_name} SET LastUpdateDt = datetime('now'), EndDt = datetime('now'), Status = ?, Result = ? WHERE {self.task_name}Id = ?", (status, result, work_item_id))
        elif status in ['RUNNING']:
            self.db.execute(f"UPDATE {self.queue_name} SET LastUpdateDt = datetime('now'), StartDt = datetime('now'), Status = ?, Result = ? WHERE {self.task_name}Id = ?", (status, result, work_item_id))
        else:
            self.db.execute(f"UPDATE {self.queue_name} SET LastUpdateDt = datetime('now'), Status = ?, Result = ? WHERE {self.task_name}Id = ?", (status, result, work_item_id))
        self.db.commit()



import builtins
import collections
import collections.abc
import csv
import datetime
import decimal
import fnmatch
import functools
import hashlib
import itertools
import json
import logging
import numbers
import os
import random
import re
import socket
import sys
import time
import traceback
import typing
import uuid
import zoneinfo


# Configuration helpers

def get_config(key, default_value=None):
    if default_value is None:
        assert f"PMG_{key}" in os.environ, f"Configuration key {key} (environment variable PMG_{key}) not found."
    return os.environ.get(f"PMG_{key}", default_value)

def get_config_keys(pattern):
    return {k[4:]: v for k, v in os.environ.items() if k.startswith('PMG_') and fnmatch.fnmatch(k[4:], pattern)}


# Global configuration variables

ASCII_SUB = '\x1a'
DEFAULT_CSV_DELIMITER = '\t'
DEFAULT_HASH_ALGO = hashlib.sha1
TRACE_LOG = get_config('TRACE_LOG', '0') == '1'

# Object Helpers

class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

def coalesce(*arg):
    for el in arg:
        if el is not None:
            return el
    return None


# ID Helpers

def newid():
    return str(uuid.uuid4())

def newrunid():
    return f"{utcnow().strftime('%y%m%d%H%M%S%f')}{''.join([random.choice('ABCDEFGHJKLMNPQRSTUVWXYZ') for i in range(4)])}"


# Collection Helpers

def empty(obj):
    return obj is None or len(obj) == 0

def blank(obj):
    return obj is None or str(obj).strip() == ''

def allblank(obj):
    return all(map(blank, obj))


# Formatting helpers

def to_sql(obj):
    if obj is None:
        return 'NULL'
    if isinstance(obj, str):
        return "'{}'".format(obj.replace("'", "''"))
    if isinstance(obj, bool):
        return '1' if obj else '0'
    # datetime.datetime inherits from datetime.date and will be caught
    if isinstance(obj, (datetime.date, datetime.time)):
        return f"'{obj.isoformat()}'"
    if isinstance(obj, numbers.Real):
        return str(obj)
    assert False, 'We should not be here.'

def to_csv(obj, delim='\t', delim_sub: str = None,
           float_decimals: int = 10, bool_true_value: str = None, bool_false_value: str = None):
    s = ''
    delim_sub = coalesce(delim_sub, ASCII_SUB)
    bool_true_value = coalesce(bool_true_value, '1')
    bool_false_value = coalesce(bool_false_value, '0')
    if obj is not None:
        # datetime.datetime inherits from datetime.date and will be caught
        if isinstance(obj, (datetime.date, datetime.time)):
            return obj.isoformat()
        if isinstance(obj, numbers.Real):
            s = str(round(obj, float_decimals))
        elif isinstance(obj, bool):
            s = bool_true_value if obj else bool_false_value
        else:
            s = str(obj)
    assert delim_sub not in s, 'Cannot safely serialize data that contains ASCII_SUB character'
    return s.replace(delim, delim_sub)

def to_namedtuple(dictionary: dict, tuple_name: str = None):
    return collections.namedtuple(tuple_name or 'Record', dictionary.keys())(**dictionary)

def make_iterable(obj):
    if isinstance(obj, str) or not isinstance(obj, collections.abc.Iterable):
        return [obj]
    return obj


# File Helpers

if os.name != 'nt':
    import pwd
    import grp

def touch(fname, mode=0o666, dir_fd=None, **kwargs):
    flags = os.O_CREAT | os.O_APPEND
    with os.fdopen(os.open(fname, flags=flags, mode=mode, dir_fd=dir_fd)) as f:
        os.utime(f.fileno() if os.utime in os.supports_fd else fname,
                 dir_fd=None if os.supports_fd else dir_fd, **kwargs)

def chownmod(path, owner=None, group=None, perms=None):
    assert os.name != 'nt', "chownmod() is not available in Windows, sadly."
    if owner is not None or group is not None:
        uid = pwd.getpwnam(owner).pw_uid if owner is not None else -1
        gid = grp.getgrnam(group).gr_gid if group is not None else -1
        os.chown(path, uid, gid)
    if perms is not None:
        os.chmod(path, perms)

def getfiles(path):
    for root, _, files in os.walk(path):
        for fn in files:
            yield os.path.join(root, fn)

def iterfiles(path):
    for fn in os.listdir(path):
        fullpath = os.path.join(path, fn)
        if os.path.isdir(fullpath):
            continue
        yield fullpath

def iterdirs(path):
    for fn in os.listdir(path):
        fullpath = os.path.join(path, fn)
        if not os.path.isdir(fullpath):
            continue
        yield fullpath

def hashfile(path, hash_algo=None):
    h = coalesce(hash_algo, DEFAULT_HASH_ALGO())
    with open(path, 'rb') as f:
        h.update(f.read())
    return h.hexdigest()

def writetextfile(path, content):
    with open(path, 'wt', encoding='utf-8', newline='') as g:
        g.write(content)

# String Helpers

def mass_replace(s, replace_vars):
    if s is None:
        return s
    if not isinstance(replace_vars, dict):
        raise Exception('replace_vars must be a dict.')
    for v in replace_vars.keys():
        s = s.replace(v, replace_vars[v])
    return s

def string_to_dict(records_string: str,
                   record_delim: str = ',',
                   pair_delim: str = '=',
                   key_processing_function: typing.Callable = None) -> dict:
    return {coalesce(key_processing_function, identity)(kvp[0]): kvp[1]
            for kvp in (kvp.split(pair_delim)
                        for kvp in records_string.split(record_delim))}

def remove_prefix(text, prefix):
    return text[text.startswith(prefix) and len(prefix):]

def remove_suffix(text, suffix):
    return text[-(text.endswith(suffix) and len(suffix)):]

def maxlen(obj, max_len, ellipsis=None):
    if len(obj) <= max_len:
        return obj
    if ellipsis is None:
        return obj[:max_len]
    return f"{obj[:max_len]}{ellipsis}"

def hide_str_excess(text, max_len):
    return maxlen(text, max_len, '...')

def hashstr(s, hash_algo=None):
    h = coalesce(hash_algo, DEFAULT_HASH_ALGO())
    if isinstance(s, bytes):
        h.update(s)
    else:
        h.update(s.encode('utf-8'))
    return h.hexdigest()

def camel_case_split(identifier):
    matches = re.finditer('.+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)', identifier)
    return [m.group(0) for m in matches]

def quote(s, quote_char):
    assert quote_char not in s, f"String '{s}' already contains quote ({quote_char}), escaping is needed."
    return f"{quote_char}{s}{quote_char}"

def unquote(s, quote_char):
    if len(s) >= 2 * len(quote_char) and s[:len(quote_char)] == quote_char and s[-len(quote_char):] == quote_char:
        return s[len(quote_char):-len(quote_char)]
    return s

def strif(cond, s):
    return s if cond else ''

# Date/Time Helpers

def today(date_format: str = None) -> str:
    """Return today's date as an ISO formatted string or with the provided format."""
    if date_format is None:
        return datetime.datetime.today().date().isoformat()
    return datetime.datetime.today().date().strftime(date_format)

def now(date_format: str = None, tz_name: str = None) -> str:
    dt = None
    if timezone is None:
        dt = datetime.datetime.now()
    else:
        dt = datetime.datetime.now(zoneinfo.ZoneInfo(tz_name))
    if date_format is None:
        return dt.isoformat()
    return dt.strftime(date_format)

def timezone(tz_name):
    return zoneinfo.ZoneInfo(tz_name)

def utcnow():
    return datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)

def localnow(tz_name):
    return datetime.datetime.now(tz=timezone(tz_name))

def is_datetime_naive(dt):
    return dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None

def is_datetime_utc(dt):
    return not is_datetime_naive(dt) and \
        (dt.tzinfo == datetime.timezone.utc or
         dt.tzinfo.utcoffset(dt).total_seconds() == 0)

def make_datetime_naive(dt):
    return datetime.datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond)

def parse_iso_date(dt):
    d = [int(date_part) for date_part in dt.split('-')]
    return datetime.date(d[0], d[1], d[2])

def parse_iso_datetime_utc(dt):
    assert dt[4] == '-' and dt[7] == '-' and (dt[10] == 'T' or dt[10] == ' ') and dt[13] == ':' and dt[16] == ':', 'Unexpected ISO date/time format'
    if dt[-6] in '+-':
        assert dt[-5:] == '00:00', 'Timezone must be UTC / offset 0:00'
        dt = dt[:-6]
    if len(dt) > 19 and dt[19] == '.':
        return datetime.datetime(int(dt[0:4]), int(dt[5:7]), int(dt[8:10]), int(dt[11:13]), int(dt[14:16]), int(dt[17:19]), int(dt[20:]), tzinfo=datetime.timezone.utc)
    return datetime.datetime(int(dt[0:4]), int(dt[5:7]), int(dt[8:10]), int(dt[11:13]), int(dt[14:16]), int(dt[17:19]), tzinfo=datetime.timezone.utc)

def get_next_weekday(weekday, from_date=None):
    """Get the next date of a particular weekday from the from_date (default: today).
       from_date is included as a valid return date (e.g., if you ask for the next Monday and today is Monday,
       you get today's date. If you dislike this, give from_date as tomorrow.)
    """
    if isinstance(from_date, datetime.datetime):
        from_date = from_date.date()
    else:
        from_date = coalesce(from_date, datetime.datetime.now().date())
    return from_date + datetime.timedelta((weekday - from_date.weekday()) % 7)

def get_seconds_until(target_dt, from_dt=None):
    return (target_dt - coalesce(from_dt, utcnow())).total_seconds()


# System Helpers

def get_exc_info():
    return '\n'.join(traceback.format_exception(*sys.exc_info()))


# CSV Helpers
def csv_to_namedtuples(path, record_type: type = None, **kwargs):
    Record = None
    record_type = coalesce(record_type, collections.namedtuple)
    if 'delimiter' not in kwargs:
        kwargs.update({'delimiter': DEFAULT_CSV_DELIMITER})
    with open(path, 'rt') as f:
        reader = csv.reader(f, **kwargs)
        for i, r in enumerate(reader):
            if i == 0:
                Record = record_type('Record', r)
                continue
            yield Record(*r)


# JSON Helpers

class FriendlyJsonEncoder(json.JSONEncoder):
    def default(self, o):
        try:
            if isinstance(o, decimal.Decimal):
                return format(o, 'f')
            if isinstance(o, (datetime.date, datetime.time)):
                return o.isoformat()
            if isinstance(o, bytes):
                try:
                    return o.decode('utf-8')
                except UnicodeError:
                    return f'<Binary ({len(o)} bytes)>'
            if isinstance(o, set):
                return tuple(o)
            return super(FriendlyJsonEncoder, self).default(o)
        except Exception as ex:
            return f'<Unencodable type {builtins.type(o)} ({type(ex)})>'

def to_json(obj, **kwargs):
    return json.dumps(obj, cls=FriendlyJsonEncoder, **kwargs)

def pretty_json(js):
    return json.dumps(js, indent=4)


# Logging Helpers

LOG_CONTEXT = {'host': socket.gethostname()}
class JsonFormatter(logging.Formatter):
    def formatException(self, exc_info):
        result = super().formatException(exc_info)
        return repr(result)

    def format(self, record):
        super().format(record)
        js = {}
        for k, v in record.__dict__.items():
            if k not in 'msg args levelno filename exc_info created msecs relativeCreated thread threadName processName process':
                if v:
                    js[k] = v
            js['created'] = datetime.datetime.fromtimestamp(record.created)
            js.update(LOG_CONTEXT)
        return json.dumps(js, cls=FriendlyJsonEncoder)

def configure_logging():
    handler = logging.StreamHandler()
    formatter = JsonFormatter(logging.BASIC_FORMAT)
    handler.setFormatter(formatter)
    root = logging.getLogger()
    root.setLevel(os.environ.get('PMG_LOGLEVEL', 'INFO'))
    root.addHandler(handler)

def logwrap(logger, swallow_exception=False):
    def innerlogwrap(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            correlation_id = newid()
            start_time = time.process_time()
            LOG_CONTEXT['correlation_id'] = correlation_id
            if TRACE_LOG:
                logger.debug(f'> Begin function {func.__name__}()', extra={'function': {'name': func.__name__, 'args': args, **kwargs}})
            try:
                res = func(*args, **kwargs)
            except Exception as e:
                logger.exception(f'Exception in function {func.__name__}()', extra={'function': {'name': func.__name__, 'args': args, **kwargs}}, exc_info=e)
                if not swallow_exception:
                    raise
            if TRACE_LOG:
                logger.debug(f'> End function {func.__name__}() after {time.process_time() - start_time:,.3f} seconds')
            return res
        return wrapper
    return innerlogwrap

def is_weekend(dt):
    return dt.weekday() in [5, 6]

def identity(s):
    assert s == s, 'Invalid value for identity function'
    return s

def skip(iterable, entries_to_skip):
    return itertools.islice(iterable, entries_to_skip, None)

class UserError(Exception):
    pass



"""LaTeX-related helper functions"""

from . import mass_replace

LATEX_ESCAPE = {
    '\\': "\\backslash",
    '&': '\\&',
    '%': '\\%',
    '$': '\\$',
    '#': '\\#',
    '_': '\\_',
    '{': '\\{',
    '}': '\\}',
    '~': '\\textasciitilde',
    '^': '\\textasciicircum',
    '\n': ' \\newline '
}

def escape(s):
    return mass_replace(s, LATEX_ESCAPE)



from hypothesis import given
import hypothesis.strategies as st
import pmg.latex

@given(st.text())
def test_latex_escape(text):
    escaped_text = pmg.latex.escape(text)
    assert '~' not in escaped_text
    assert '^' not in escaped_text



import os
import shutil
import pmg.temp

def test_tempdir():
    with pmg.temp.tempdir() as path:
        assert os.path.isdir(path)
    assert not os.path.isdir(path)
    with pmg.temp.tempdir(remove_upon_completion=False) as path:
        assert os.path.isdir(path)
    assert os.path.isdir(path)
    shutil.rmtree(path)
    assert not os.path.isdir(path)



"""Helper functions to work with DBAPI compliant databases and connection strings"""

from . import to_namedtuple, string_to_dict

def connection_string_to_dict(connection_string: str) -> dict:
    return string_to_dict(connection_string, ';', '=', lambda s: s.upper())

def execute(cn, sql, *args, **kwargs):
    cur = cn.cursor()
    cur.execute(sql, *args, **kwargs)
    return cur

def iter_cursor_as_dict(cur):
    columns = [column[0] for column in cur.description]
    for r in cur:
        yield dict(zip(columns, r))

def iter_cursor_as_namedtuple(cur, tuple_name: str = None):
    for r in iter_cursor_as_dict(cur):
        yield to_namedtuple(r, tuple_name)

def iter_cursor_sets(cur):
    "Iterate through all sets provided by a Python DB API 2.0 cursor"
    more = True
    if cur.description is not None:
        yield cur
    while more:
        more = cur.nextset()
        if cur.description is not None:
            yield cur

def iter_named_cursor_sets(cur):
    """Iterate through all named sets provided by a Python DB API 2.0 cursor.
    A set is named by the preceding set providing a single scalar called
    _DataTableName.
    """
    tab = None
    for rs in iter_cursor_sets(cur):
        if tab is None:
            assert len(rs.description) == 1 and \
                   rs.description[0][0] == '_DataTableName', \
                   'Every other result must be a scalar named _DataTableName.'
            tab = rs.fetchone()[0]
        else:
            yield (tab, rs)
            tab = None



import random
from hypothesis import given
import hypothesis.strategies as st
import pmg.db
import sqlite3
import sys

@given(st.text(alphabet=st.characters(blacklist_characters=';='), min_size=1),
       st.text(alphabet=st.characters(blacklist_characters=';='), min_size=1),
       st.text(alphabet=st.characters(blacklist_characters=';='), min_size=1),
       st.text(alphabet=st.characters(blacklist_characters=';='), min_size=1))
def test_connection_string_to_dict(server, database, user, password):
    cn = pmg.db.connection_string_to_dict("DRIVER={ODBC Driver 17 for SQL Server};"
                                          f"Server={server};Database={database};"
                                          f"UID={user};PWD={password}")
    assert cn['DRIVER'] == '{ODBC Driver 17 for SQL Server}'
    assert cn['SERVER'] == server
    assert cn['DATABASE'] == database
    assert cn['UID'] == user
    assert cn['PWD'] == password

def init_test_db(db):
    db.execute("CREATE TABLE IF NOT EXISTS TestTable (IntValue INTEGER NOT NULL, TextValue TEXT NULL);")
    db.executemany("INSERT INTO TestTable VALUES (?, ?)", ((random.randint(-sys.maxsize + 1, sys.maxsize), random.random()) for i in range(100)))
    db.commit()

def test_iter_cursor():
    with sqlite3.connect(':memory:') as db:
        init_test_db(db)
        cur = db.execute("SELECT * FROM TestTable")
        for r in pmg.db.iter_cursor_as_dict(cur):
            assert 'TextValue' in r and 'IntValue' in r
        cur = db.execute("SELECT * FROM TestTable")
        for r in pmg.db.iter_cursor_as_namedtuple(cur):
            assert isinstance(r.TextValue, str)
            assert isinstance(r.IntValue, int)



"""Helper functions to work with temporary storage"""

import contextlib
import os
import shutil
import tempfile

@contextlib.contextmanager
def cd(newdir, cleanup=lambda: True):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)
        cleanup()

@contextlib.contextmanager
def tempdir(remove_upon_completion=True):
    dirpath = tempfile.mkdtemp()

    def cleanup():
        shutil.rmtree(dirpath)

    if remove_upon_completion:
        with cd(dirpath, cleanup):
            yield dirpath
    else:
        with cd(dirpath):
            yield dirpath
